/**
 * Constants.
 */
const NS_LOCAL_ANNOTATIONS_SERVER_CID = Components.ID("{32685ae0-1c9d-4862-b022-4f59dd3fa8b6}");
const NS_LOCAL_ANNOTATIONS_SERVER_PROG_ID = "@mozilla.org/local_annotations_server;1";


/**
 * Global vars
 */

// global RDF Service
var gRDFService = Components
    .classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);

/**
 * A local annotations server, storing its data
 * in RDF and implementing
 * Components.interfaces.nsIAnnotationsServer
 */
function LocalServer() {

    // whether the datasource is loaded
    this.loaded = false;

    var that = this; /* Copy for use inside closure */
    
    // observer to observer loading of datasource
    var localLoadObserver = {
        onBeginLoad: function(aSink) {},
        onInterrupt: function(aSink) {},
        onResume   : function(aSink) {},
        onEndLoad  : function(aSink)
	{ 
	    that.loaded = true;
	    dump("Local RDFDataSource loaded.\n");
	},
        onError    : function(aSink, aStatus, aErrorMsg) { 
	    dump("Error loading LocalServer dataSource, " + aErrorMsg + "\n"); 
	}
    };
    
    try {
	// get the preferences directory
	var fileLocator = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);  
	var file = fileLocator.get("ProfD", Components.interfaces.nsIFile);
	
	// append servers filename
	file.append(LocalServer.LOCAL_ANNOTATIONS_FILENAME);	
	
	// if the file does not exist, create it
	if (!file.exists()) {
	    file = LocalServer.createNewAnnotationsFile(file);
	}		
	
	// get the data source
	this.dataSource = gRDFService.GetDataSource(LocalServer.getURLSpecFromFile(file));
	
	// test whether it's loaded
	var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
	if (remote.loaded) {
	    dump("LocalServer: datasource already loaded!\n");
	    this.loaded = true;
        } 
	else {
            // Attach an observer to the loading of the RDF
            var sink =
                this.dataSource.QueryInterface(Components.interfaces.nsIRDFXMLSink);
            sink.addXMLSinkObserver(localLoadObserver);
        }	
    }
    catch(e) {
	dump("LocalServer Loading Error: " + e + "\n");
    }
}

LocalServer.LOCAL_ANNOTATIONS_FILENAME = "localAnnotations.rdf";
LocalServer.EMPTY_ANNOTATIONS_RDF = 
    "<?xml version=\"1.0\">\n" +
    "<r:RDF \n"+
    "xmlns:d=\"http://purl.org/dc/elements/1.0/\" \n" +
    "xmlns:t=\"http://www.w3.org/2001/03/thread#\" \n" +
    "xmlns:atr=\"http://www.w3.org/2001/12/attributions/ns#\" \n" +
    "xmlns:http=\"http://www.w3.org/1999/xx/http#\" \n" +
    "xmlns:p=\"xmlns:http://www.w3.org/2000/08/palm56/addr#\" \n" +
    "xmlns:a=\"http://www.w3.org/2000/10/annotation-ns#\" \n" +
    "xmlns:r=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n</r:RDF>\n";

// File read/write operations
LocalServer.NS_RDONLY      = 0x01;
LocalServer.NS_WRONLY      = 0x02;
LocalServer.NS_CREATE_FILE = 0x08;
LocalServer.NS_TRUNCATE    = 0x20;

// user -- read, write; group -- read; other -- read
LocalServer.DEFAULT_FILE_PERMS = 00400 | 00200 | 00040 | 00004;


LocalServer.URI = "urn:localserver";
LocalServer.USERNAME = "noUser";
LocalServer.PASSWORD = "noPass";


/**
 * Arbitrary namespace for the body of an Annotation.
 */
LocalServer.LOCALSERVER_NS = "http://cnx.rice.edu/annotations/localServer#";
LocalServer.ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";
LocalServer.ANNOT_TYPE_NS = "http://www.w3.org/2000/10/annotationType#";
LocalServer.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
LocalServer.DC_NS    = "http://purl.org/dc/elements/1.0/";
LocalServer.THREAD_NS= "http://www.w3.org/2001/03/thread#";
LocalServer.HTTP_NS = "http://www.w3.org/1999/xx/http#";


/**
 * Literals to parse from RDF.
 */
LocalServer.REQUIRED_METADATA_LITERALS = [   
    { namespace: LocalServer.DC_NS, property: "creator"},
    { namespace: LocalServer.ANNOT_NS, property: "created"},
    { namespace: LocalServer.DC_NS, property: "date"},
    { namespace: LocalServer.ANNOT_NS, property: "context"} ];

LocalServer.REQUIRED_METADATA_RESOURCES = [
    { namespace: LocalServer.RDF_NS, property: "type"},
    { namespace: LocalServer.ANNOT_NS, property: "annotates"} ];


/**
 * RDF Parsing for Annotation namespace.
 */
LocalServer.ANNOTATION_TYPE = LocalServer.ANNOT_NS + "Annotation";

/**
 * RDF Parsing for Reply namespace.
 */
LocalServer.REPLY_TYPE = LocalServer.THREAD_NS + "Reply";


LocalServer.prototype = {
    
    get url() { return LocalServer.URI; },

    get username() { return LocalServer.USERNAME; },
    set username(name) { throw Components.Exception("You cannot set the LocalServer username."); },

    get password() { return LocalServer.PASSWORD; },
    set password(name) { throw Components.Exception("You cannot set the LocalServer password.");  }
};

LocalServer.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.nsIAnnotationsServer) &&
	!iid.equals(Components.interfaces.nsISupports)) {
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}
    return this;
}

/**
 * No extra initialization necessary for local server.
 */
LocalServer.prototype.init = function(url, username, password) {}

/**
 * Post a new annotation to the server.  
 *
 * @param data nsIMetadata for a new annotation 
 * @param bodyText string text of annotation body
 * @param observer 
 */
LocalServer.prototype.postAnnotation = function(data, bodyText, observer) {
    var annotationUriDigitsLength = 10;

    if (! this.loaded) {
	observer.postAnnotationOnFailure("LocalServer RDF Data Source still loading.  Please try again momentarily.");
	return;
    }

    /* ensure that the metadata argument has the required fields,
       and load them in */
    var unionObj = null;
    try {
	unionObj = LocalServer.loadPostData(data);
    }
    catch(e) {
	observer.postAnnotationOnFailure("Insufficient data to post annotation: " + e.toString());
	return;
    }

    // extract literals and resources from union object
    var literals = unionObj.lit;
    var resources = unionObj.res;
	
    var annotation = Components.classes["@mozilla.org/annotation;1"].createInstance(Components.interfaces.nsIAnnotation);
    var bareNumbers = LocalServer.generateRandomNumberString(annotationUriDigitsLength);
    var annotURI = "urn:annot" + bareNumbers;
    var bodyURI = "urn:body" + bareNumbers;
	
    // init the annotation
    annotation.initialize(LocalServer.URI, annotURI, bodyURI);

    var annotResource = gRDFService.GetResource(annotURI);

    // get the url that we're annotating
    var annotates = resources[LocalServer.ANNOT_NS + "annotates"].value;
    
    // assert our annotation under this url
    this.dataSource.Assert(gRDFService.GetResource(annotates),
			   gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy"),
			   annotResource, true);
	
    // assert the resources
    for(prop in resources) {
	var resource = resources[prop];
	this.dataSource.Assert(annotResource,
			       gRDFService.GetResource(resource.namespace + resource.property),
			       gRDFService.GetResource(resource.value), true);
	annotation.data.setResourceProperty(resource.namespace, resource.property, 
					    resource.value);
    }
    
    // assert the literals
    for (prop in literals) {
	var literal = literals[prop];
	this.dataSource.Assert(annotResource,
			       gRDFService.GetResource(literal.namespace + literal.property),
			       gRDFService.GetLiteral(literal.value), true);
	annotation.data.setLiteralProperty(literal.namespace, literal.property, 
					    literal.value);
    }
    
    // assert the body
    var bodyResource = gRDFService.GetResource(bodyURI);
    
    // if there is no bodyText, use empty string
    if (! bodyText) {
	bodyText = "";
    }
    this.dataSource.Assert(annotResource,
			   gRDFService.GetResource(LocalServer.ANNOT_NS + "body"),
			   bodyResource, true);
    this.dataSource.Assert(bodyResource,
			   gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "bodyText"),
			   gRDFService.GetLiteral(bodyText), true);		   

    // set the bodyText, for efficiency
    annotation.setBodyText(bodyText);

    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();
	
    observer.postAnnotationOnSuccess(annotation);
}


/**
 * Load the annotations on this server for the given URI.
 *
 * @param uri string URI to load annotations for
 * @param observer 
 */
LocalServer.prototype.loadAnnotations = function(uri, observer) {
    
    if (! this.loaded) {
	observer.loadAnnotationsOnFailure(LocalServer.URI, uri, 
					  "LocalServer RDF Data Source still loading.  Please try again momentarily.");
    }
    else {
	var annotations = this._loadAnnotationsForURI(uri);
	observer.loadAnnotationsOnSuccess(LocalServer.URI, uri, 
					  annotations.length, annotations);
    }
}
	
/**
 * Auto-load the annotations on this server for the given URI.
 *
 * @param uri string URI to autoload annotations for
 * @param observers nsISimpleEnumerator of observers 
 */					       
LocalServer.prototype.autoLoadAnnotations = function(uri, observers) {
    if (! this.loaded) {
	while (observers.hasMoreElements()) {
	    var observer = observers.getNext().QueryInterface(nsIAnnotationsObserver);
	    observer.autoLoadFailureForServer(LocalServer.URI, uri, 
					      "LocalServer RDF Data Source still loading.  Please try again momentarily.");
	}
    }
    else {
	var annotations = this._loadAnnotationsForURI(uri);
	while (observers.hasMoreElements()) {
	  var observer = observers.getNext().QueryInterface(Components.interfaces.nsIAnnotationsObserver);
	  // clone the annotations to ensure that they don't get corrupted by an earlier observer
	  observer.autoLoadAnnotationsForServer(LocalServer.URI, uri,
						annotations.length, 
						LocalServer._cloneAnnotations(annotations));
	}
    }
}

/**
 * Cancel a pending autoLoad request, if one exists.
 * For LocalServer, this is a nop.
 */
LocalServer.prototype.cancelAutoLoadRequest = function() {}


/**
 * Delete the annotation at the specified uri.
 *
 * @param uri string uri to delete
 * @param observer 
 */
LocalServer.prototype.deleteAnnotation = function(annotURI, observer) {
    var annotates = null;

    if (! this.loaded) {
	observer.deleteAnnotationOnFailure(annotURI, "LocalServer RDF Data Source still loading.  Please try again momentarily.");
	return;
    }

    // get URIs and Resources for Annotation and its Body
    // relies on uris being urn:annot<num> and urn:body<num>
    var bodyURI = annotURI.replace("annot", "body");
    var annotResource = gRDFService.GetResource(annotURI);
    var bodyResource = gRDFService.GetResource(bodyURI);

    // unassert the body text
    var bodyTextPropertyResource = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + 
							   "bodyText");
    var bodyTextValueNode = this.dataSource.GetTarget(bodyResource, bodyTextPropertyResource, true);
    this.dataSource.Unassert(bodyResource, bodyTextPropertyResource, bodyTextValueNode);

    // unassert all arcs from annotation
    var arcs = this.dataSource.ArcLabelsOut(annotResource);
    while (arcs.hasMoreElements()) {
	var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
	var arcValueNode = this.dataSource.GetTarget(annotResource, arc, true);
	// if this is annotates, store it
	if (arc.Value == LocalServer.ANNOT_NS + "annotates") {
	    annotates = arcValueNode.QueryInterface(Components.interfaces.nsIRDFResource).Value;
	}
	this.dataSource.Unassert(annotResource, arc, arcValueNode);
    }

    // unassert the annotatedBy property
    if (! annotates) {
	throw Components.Exception("Malformed RDF: couldn't get annotates property");
    }
    var uriResource = gRDFService.GetResource(annotates);
    var annotatedByProperty = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy");
    this.dataSource.Unassert(uriResource, annotatedByProperty, annotResource);

    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();
	
    observer.deleteAnnotationOnSuccess(annotURI);
}

/**
 * Update the annotation passed in on the server.  
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
LocalServer.prototype.updateAnnotation = function(annotation, observer) {
    var oldAnnotates = null;

    // ensure that annotationURI is properly defined
    if ( (!annotation.annotationURI) || (annotation.annotationURI == "") ) {
	observer.updateAnnotationOnFailure("", "annotationURI for parameter annotation is not defined.");
	return;
    }

    // ensure that bodyURI is properly defined
    if ( (!annotation.bodyURI) || (annotation.bodyURI == "") ) {
	observer.updateAnnotationOnFailure(annotation.annotationURI, "bodyURI for parameter annotation is not defined.");
	return;
    }

    // get URIs and Resources for Annotation and its Body
    var annotResource = gRDFService.GetResource(annotation.annotationURI);
    var bodyResource = gRDFService.GetResource(annotation.bodyURI);
    
    /* ensure that the metadata argument has the required fields,
       and load them in */
    var unionObj = null;
    try {
	unionObj = LocalServer.loadPostData(annotation.data);
    }
    catch(e) {
	observer.updateAnnotationOnFailure("Insufficient data to update annotation: " + e.toString());
	return;
    }

    // extract literals and resources from union object
    var literals = unionObj.lit;
    var resources = unionObj.res;
	
    // unassert all the old arcs out of the annotation except body
    var arcs = this.dataSource.ArcLabelsOut(annotResource);
    while(arcs.hasMoreElements()) {
	var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
	var arcValueNode = this.dataSource.GetTarget(annotResource, arc, true);
	// if this is annotates, store it
	if (arc.Value == LocalServer.ANNOT_NS + "annotates") {
	  oldAnnotates = arcValueNode.QueryInterface(Components.interfaces.nsIRDFResource).Value;
	}
	else if (arc.Value == (LocalServer.ANNOT_NS + "body")) {
	  continue;
	}
	this.dataSource.Unassert(annotResource, arc, arcValueNode);
    }

    // assert the new resources
    for(prop in resources) {
	var resource = resources[prop];
	
	this.dataSource.Assert(annotResource,
			       gRDFService.GetResource(resource.namespace + resource.property),
			       gRDFService.GetResource(resource.value), true);
    }
    
    // assert the new literals
    for (prop in literals) {
	var literal = literals[prop];
	this.dataSource.Assert(annotResource,
			       gRDFService.GetResource(literal.namespace + literal.property),
			       gRDFService.GetLiteral(literal.value), true);
    }
    	
    // check whether we need to change the annotatedBy property
    if (! oldAnnotates) {
      throw Components.Exception("Malformed RDF: couldn't get annotates property");
    }
    var newAnnotates = annotation.data.getProperty(LocalServer.ANNOT_NS, "annotates");
    if (oldAnnotates != newAnnotates) {
	// unassert the old annotates
	var oldResource = gRDFService.GetResource(oldAnnotates);
	var annotatedByProperty = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy");
	this.dataSource.Unassert(oldResource, annotatedByProperty, annotResource);

	// assert the new annotates
	var newResource = gRDFService.GetResource(newAnnotates);
	this.dataSource.Assert(newResource, annotatedByProperty, annotResource, true);
    }

    // do the body if necessary
    var bodyText = annotation.getBodyText();
    if (bodyText) {
	var bodyTextPropertyResource = gRDFService.GetResource(LocalServer.LOCALSERVER_NS + 
							       "bodyText");
	var bodyTextValueNode = this.dataSource.GetTarget(bodyResource, 
							  bodyTextPropertyResource, true);
	this.dataSource.Change(bodyResource, bodyTextPropertyResource, bodyTextValueNode,
			       gRDFService.GetLiteral(bodyText));

    }
    
    // flush the data source
    var remote = this.dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
    remote.Flush();
	
    observer.updateAnnotationOnSuccess(annotation);
}

    
    
     
    


/**
 * Load the replies for an annotation.
 *
 * @param annotation nsIAnnotation
 * @param observer nsIAnnotationsServerObserver
 */
LocalServer.prototype.getReplies = function(annotation, observer) {
    observer.getRepliesOnFailure(annotation, "getReplies will not be implemented until the W3C releases the necessary protocols");
}

/**
 * Reply to an annotation.
 *
 * @param original nsIAnnotation original annotation
 * @param reply nsIMetadata reply data
 * @param observer nsIAnnotationsServerObserver observer
 */
LocalServer.prototype.replyToAnnotation = function(original, reply, observer) {
    observer.replyToAnnotationOnFailure(original, "replyToAnnotation will not be implemented until the W3C releases the necessary protocols");
}

/**
 * Load the body of an annotation into the "body" Metadata
 * field of the annotation.
 *
 * @param annotation nsIAnnotation to load body for
 * @param observer nsIAnnotationsObserver
 */
LocalServer.prototype.loadAnnotationBody = function(annotation, observer) {
    if ( (!annotation.bodyURI) || (annotation.bodyURI == "") ) {
	observer.loadAnnotationBodyOnFailure(annotation, 
					     "Annotation bodyURI is not defined");
	return;
    }
    var bodyText = this.getAnnotationBodyText(annotation);
    
    if (! bodyText) {
	observer.loadAnnotationBodyOnFailure(annotation,
					     "Unable to find body text in local annotations RDF file.");
	return;
    }
    
    annotation.setBodyText(bodyText);
    observer.loadAnnotationBodyOnSuccess(annotation, bodyText);
}
    
    

/**
 * toString.
 *
 * @return String rep of this
 */
LocalServer.prototype.toString = function() { return "[LocalServer]"; }


// PRIVATE methods

LocalServer.createNewAnnotationsFile = function(file) {
     
    // initialize local file
    var localFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
    
    localFile.initWithPath(file.path);
    
    if (localFile.exists() && 
	(localFile.isDirectory() || !localFile.isWritable())) {
	throw Components.results.NS_ERROR_FAILURE;		
    }
	
    // write flags
    var flags = LocalServer.NS_WRONLY | LocalServer.NS_CREATE_FILE | LocalServer.NS_TRUNCATE;

    // get a file output stream
    var fileStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
    fileStream.init(localFile, flags, LocalServer.DEFAULT_FILE_PERMS, 0);

    fileStream.write(LocalServer.EMPTY_ANNOTATIONS_RDF, LocalServer.EMPTY_ANNOTATIONS_RDF.length);
    fileStream.close();

    return localFile;
}

/**
 * Gets a URL spec from a file, working
 * for whichever version of mozilla you
 * happen to be running.
 *
 * @param file nsIFile to get url spec for
 * @return nsIFileURL
 */
LocalServer.getURLSpecFromFile = function(file) {
    
    if (Components.classes["@mozilla.org/network/io-service;1"] && Components.interfaces.nsIIOService) {
	
	var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
	var fileHandler = ioService.getProtocolHandler("file").QueryInterface(Components.interfaces.nsIFileProtocolHandler);

	// new version
	if (fileHandler && fileHandler.getURLSpecFromFile) {
	    return fileHandler.getURLSpecFromFile(file);
	}
	// old version in ioService
	else {
	    return ioService.getURLSpecFromFile(file);
	}
    }
    // really old version
    else {
	var file_url = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIFileURL);
	file_url.file = file;
	return file_url.spec;
    }
}


/**
 * Generates a random integral number,
 * possibly with leading zeros.
 *
 * @param length length of number
 */
LocalServer.generateRandomNumberString = function(length) {
    var uri = "";

    for (var i=0; i < length; i++) {
	// get an integer of uri_length digits
	rand = Math.floor(Math.random() * 10);
	uri += rand;
    }
    
    return uri;
}

/**
 * Loads the values from an nsIAnnotationMetadata
 * object, and tests that the necessary values are defined.
 *
 * @param meta nsIAnnotationMetadata to load data from
 * @throws Exception if a required field is undefined
 * @return requiredData with values filled in if all fields exist
 */
LocalServer.loadPostData = function(meta) {
    var literals = {};
    var resources = {};
    
    // check for required literals
    for (var i = 0; i < LocalServer.REQUIRED_METADATA_LITERALS; i++) {
	var literal = LocalServer.REQUIRED_METADATA_LITERALS[i];
	var aValue = meta.getProperty(literal.namespace, literal.property);

	if (aValue) {
	    literals[literal.namespace + literal.property] = { namespace: literal.namespace, property: literal.property, value: aValue };
	}
	else {
	    throw Components.Exception("Metadata is missing required literal " + 
				       literal.namespace + literal.property + "\n");
	}
    }
    
    // check for required resources
    for (var i = 0; i < LocalServer.REQUIRED_METADATA_RESOURCES; i++) {
	var resource = LocalServer.REQUIRED_METADATA_RESOURCES[i];
	var aValue = meta.getProperty(resource.namespace, resource.property);

	if (aValue) {
	    resources[resource.namespace + resource.property] = 
	    { namespace: resource.namespace, property: resource.property, value: aValue };
	}
	else {
	    throw Components.Exception("Metadata is missing required resource " + 
				       resource.namespace + resource.property + "\n");
	}
    }

    // enumerate the metadata to add additional items
    var properties = {};
    var count = {};
    meta.getAllProperties(count, properties);
    for (var i = 0; i < properties.value.length; i++) {
	var datum = properties.value[i];
	if (datum.type == datum.LITERAL) {
	    if (! literals[datum.namespace + datum.property]) {
		
		literals[datum.namespace + datum.property] = 
		{ namespace: datum.namespace, property: datum.property, value: datum.value };
	    }
	}
	else {
	    if (! resources[datum.namespace + datum.property]) {
		resources[datum.namespace + datum.property] = 
		{ namespace: datum.namespace, property: datum.property, value: datum.value };
	    }
	}
    }

    return { lit: literals, res: resources };
}


/**
 * Given an rdfDataSource, a resourceEnum (subset of the dataSource) and the resources and literals to extract,
 * creates <code>Annotation<code>s with these resources and literals in the metadata
 * of the Annotation.  
 *
 * @param rdfDataSrc nsIRDFDataSource to mine for annotations
 * @param resourceEnum nsISimpleEnumerator subset of resources to build annotations for
 * @param resourceNames names of fields of this object will be parsed as resources from the RDF
 * @param literalNames names of fields of this object will be parsed as literals from the RDF
 * @param rdfType one of either ANNOTATION_TYPE or REPLY_TYPE
 * @return Annotation[] of Annotations
 */
LocalServer.prototype.buildAnnotationsFromResourceEnum = function(rdfDataSrc, resourceEnum, rdfType, serverName) 
{
    var annotations = [];

    while(resourceEnum.hasMoreElements()) {

	// get resource
        var res = resourceEnum.getNext();
        res = res.QueryInterface(Components.interfaces.nsIRDFResource);

	var annotURI = res.Value;
	var bodyURI  = LocalServer.getRDFResource(res, LocalServer.ANNOT_NS, "body", rdfDataSrc);
	
	var annotation = Components.classes["@mozilla.org/annotation;1"].createInstance(Components.interfaces.nsIAnnotation);
	annotation.initialize(serverName, annotURI, bodyURI);

	// view all the arcs out
	var arcs = rdfDataSrc.ArcLabelsOut(res);
	while (arcs.hasMoreElements()) {
	    var arc = arcs.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
	    if (arc.Value == LocalServer.ANNOT_NS + "body") {
		continue;
	    }

	    var splitValue = LocalServer.splitResourceString(arc.Value);

	    // try to get it as a resource
	    var value = LocalServer.getRDFResource(res, splitValue.namespace, splitValue.property, rdfDataSrc);
	    // if not a resource, try a literal
	    if (! value) {
		value = LocalServer.getRDFLiteral(res, splitValue.namespace, splitValue.property, rdfDataSrc);
		if (! value) {
		    throw Components.Exception("Couldn't get as literal either--" + splitValue.namespace + splitValue.property);
		}
		annotation.data.setLiteralProperty(splitValue.namespace, splitValue.property, value);
	    }
	    else {
		annotation.data.setResourceProperty(splitValue.namespace, splitValue.property, value);
	    }
	}

	// additionally, load and set the annotation bodyText
	// as an optimization
	var bodyText = this.getAnnotationBodyText(annotation);
	if (bodyText) {
	  annotation.setBodyText(bodyText);
	}

	annotations.push(annotation);
    }

    return annotations;
}


/**
 * Splits a namespace + property Resource string into
 * namespace and property.
 *
 * @param resourceString string of namespace + property
 * @return object with namespace and property fields
 */
LocalServer.splitResourceString = function(resourceString) {
    var i;
    for (i = (resourceString.length - 1); i >= 0; i--) {
	var charAtI = resourceString.charAt(i);
	if (charAtI == '#' || charAtI == '/' || charAtI == ':') {
	    return { namespace: resourceString.substring(0, i+1),
				    property: resourceString.substring(i+1) };
	}
    }
    
    // couldn't find it -- return null
    return null;
}



/**
 * Get a property from an RDF Literal.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of literal or undefined if not found
 *  
 */
LocalServer.getRDFLiteral = function(resource, ns, property, rdfDataSrc) {
    return LocalServer.getRDFValue (resource, ns, property, rdfDataSrc,
			     Components.interfaces.nsIRDFLiteral);
}

/**
 * Get a property from an RDF resource.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @return value of property, or undefined if not found
 *  
 */
LocalServer.getRDFResource = function(resource, ns, property, rdfDataSrc) {
    return LocalServer.getRDFValue (resource, ns, property, rdfDataSrc,
			     Components.interfaces.nsIRDFResource);
}

/**
 * Get an RDF Value.
 *
 * @param resource RDF resource
 * @param ns String namespace
 * @param property String property to retrieve
 * @param rdfDataSrc RDF Data Source
 * @param iface interface to query
 * @return value of resource or undefined if property not found
 *  
 */
LocalServer.getRDFValue = function(resource, ns, property, rdfDataSrc, iface) {
    try {
        var RDFNode =
	rdfDataSrc.GetTarget(resource, gRDFService.GetResource(ns + property), true);
        RDFNode = RDFNode.QueryInterface(iface);
        RDFNode = RDFNode.Value;
        return RDFNode;
    } catch (e) {
        return null;
    }
}


/**
 * Make a deep clone of an array of annotations.
 *
 * @param oldAnnots array of annotations
 * @return deep clone
 */
LocalServer._cloneAnnotations = function(oldAnnots) {
      var newAnnots = [];
      for (var i = 0; i < oldAnnots.length; i++) {
	  newAnnots.push(oldAnnots[i].clone());
      }
      return newAnnots;
}

/**
 * Loads the local annotations for a uri.
 *
 * @param uri string uri for which to load annotations
 * @return array of nsIAnnotation for uri
 */
LocalServer.prototype._loadAnnotationsForURI = function(uri) {
    
  var resourceEnum = this.dataSource.GetTargets(gRDFService.GetResource(uri),
						gRDFService.GetResource(LocalServer.LOCALSERVER_NS + "annotatedBy"),
						true);

    return this.buildAnnotationsFromResourceEnum(this.dataSource, resourceEnum,
						 LocalServer.ANNOTATION_TYPE,
						 LocalServer.URI);
}

LocalServer.prototype.getAnnotationBodyText = function(annotation) {
  return LocalServer.getRDFLiteral(gRDFService.GetResource(annotation.bodyURI),
				   LocalServer.LOCALSERVER_NS,
				   "bodyText",
				   this.dataSource);
}




/**************************
 * XPCOM stuff
 **************************/

/**
 * Factory for nsLocalAnnotationsServer.
 */
var serverFactory = new Object();

serverFactory.createInstance = function (outer, iid) {
    if (outer != null) {
	throw Components.results.NS_ERROR_NO_AGGREGATION;
    }
    
    if (!iid.equals(Components.interfaces.nsISupports)) {
	throw Components.results.NS_ERROR_INVALID_ARG;
    }

    return new LocalServer();
}


/**
 * Module for nsLocalAnnotationsServer.
 */
var nsLocalAnnotationsServerModule = new Object();

nsLocalAnnotationsServerModule.registerSelf =
    function (compMgr, fileSpec, location, type) {
	compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	dump("***** Local Annotations Server Service -- registering self ******\n");
	compMgr.registerFactoryLocation(NS_LOCAL_ANNOTATIONS_SERVER_CID,
					"Local Annotations Server Service",
					NS_LOCAL_ANNOTATIONS_SERVER_PROG_ID, 
					fileSpec,
					location,
					type);
    }

nsLocalAnnotationsServerModule.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.unregisterFactoryLocation(NS_LOCAL_ANNOTATIONS_SERVER_CID, fileSpec);
}

nsLocalAnnotationsServerModule.getClassObject = function (compMgr, cid, iid) {
    if (cid.equals(NS_LOCAL_ANNOTATIONS_SERVER_CID)) {
	return serverFactory;
    }
    else {
	throw Components.results.NS_ERROR_NO_INTERFACE;  
    }
}

nsLocalAnnotationsServerModule.canUnload = function(compMgr) {
    dump("***** unloading Local Annotations Server Service ****\n");
    return true;
}

/* entrypoint */
function NSGetModule(compMgr, fileSpec) {
    return nsLocalAnnotationsServerModule;
}



